from __future__ import division, absolute_import, print_function

import numpy as np


class Alga(object):
    def setup(self, event):
        pass

    def draw(self, time):
        """Draw objects within a 3D context"""
        pass

    def draw_shaded(self, time):
        """Draw objects within a shaded 3D context"""
        pass

    def draw2d(self, time):
        """Draw objects within a 2D context"""
        pass
